# coding: utf-8
import logging
import os
import time
import json
import cStringIO
import traceback
import copy

__author__ = 'czhou <czhou@ilegendsoft.com>'

Log = logging.getLogger("Leap")
Log.addHandler(logging.StreamHandler())
Log.setLevel(logging.DEBUG)


# coding: utf-8

import re

import ML

__author__ = 'czhou <czhou@ilegendsoft.com>'


class Role(ML.Object):
    def __init__(self, name, acl):
        super(Role, self).__init__()
        self.set_name(name)
        self.set_acl(acl)

    def get_name(self):
        """
        获取 Role 的 name，等同于 role.get('name')
        """
        return self.get('name')

    def set_name(self, name):
        """
        为 Role 设置 name，等同于 role.set('name', name)
        """
        return self.set('name', name)

    def get_users(self):
        """
        获取当前 Role 下所有绑定的用户。
        """
        return self.relation('users')

    def get_roles(self):
        return self.relation('roles')

    def validate(self, attrs):
        if 'name' in attrs and attrs['name'] != self.get_name():
            new_name = attrs['name']
            if not isinstance(new_name, basestring):
                raise TypeError('role name must be a basestring')
            r = re.compile('^[0-9a-zA-Z\-_]+$')
            if not r.match(new_name):
                raise TypeError('role\'s name can only contain alphanumeric characters, _, -, and spaces.')

        return super(Role, self).validate(attrs)



# coding: utf-8

"""MaxLeap Python SDK
"""

import client
from .acl import ACL
from .client import init, by_hook, change_region, UserPrincipal, get_master_principal
from .errors import MaxLeapError
from .file_ import File
from .geo_point import GeoPoint
from .object_ import Object
from .query import Query
from .relation import Relation
from .user import User
from .role import Role
from .server import Server, get_principal
from .log import Log
from flask import Response

__author__ = 'czhou <czhou@ilegendsoft.com>'
__version__ = '0.1.1'
DEBUG = True
PRO = False

__all__ = [
    'ACL',
    'File',
    'GeoPoint',
    'MaxLeapError',
    'Object',
    'Query',
    'Relation',
    'User',
    'client',
    'init',
    'by_hook',
    'change_region',
    'UserPrincipal',
    'get_master_principal',
    'Role',
    'Server',
    'get_principal',
    'Log',
    'Response'
]



# coding: utf-8

import os
import ML
import json
import logging
import traceback
import functools
from flask import Flask, request, Response

__author__ = 'czhou <czhou@ilegendsoft.com>'

Port = 8080
Host = '127.0.0.1'

app = Flask(__name__)

def get_principal():
    if not ML.PRO:
        return None
    str_user_principal = request.headers.get('Las-User-Principal',None)
    if not str_user_principal:
        raise RuntimeError("You can't use server sdk without UserPrincipal")
    json_user_principal = json.loads(str_user_principal)
    session_token = None
    master_key = None
    client_key = None
    if json_user_principal.get('identityType') in ["ORG_USER", "APP_USER"]:
        session_token = json_user_principal.get('sessionToken')
    if json_user_principal.get('identityType') in ["MASTER_KEY"]:
        master_key = json_user_principal.get('key')
    if json_user_principal.get('identityType') in ["API_KEY"]:
        client_key = json_user_principal.get('key')
    return ML.UserPrincipal(ML.client.APP_ID, client_key=client_key, master_key=master_key, session_token=session_token)

@app.after_request
def after_request(response):
    ML.Log.debug("===========Request End===============")
    return response

@app.before_request
def before_request():
    ML.Log.debug("===========Request Start=============")
    ML.Log.debug("Method:{0}".format(request.method))
    ML.Log.info("Headers:\n{0}".format(request.headers))
    ML.Log.info("Body:{0}".format(request.data))

@app.errorhandler(500)
def custom_error(e):
    if isinstance(e, ML.MaxLeapError):
        ML.Log.error(e)
        return Response(
            json.dumps({"errorCode":e.code,"errorMessage":e.error}),
            status=545,
            mimetype='application/json'
        )
    traces_info = traceback.format_exc()
    ML.Log.error(traces_info)
    return Response(
        traces_info,
        status=545,
        mimetype='application/json'
        )

class flask_base_server(object):
    def __init__(self, app, host, port):
        """
        创建一个新的server

        :host: IP
        :port: 端口
        :return:
        """
        self._app = app
        self._host = host
        self._port = port
        self._function_map = {}
        self._job_map = {}
        self._hook_map = {}
        self._hook_classes = set()
        self.setup()

    def setup(self):
        self.before_save = functools.partial(self.Hook, hook_name='before_save')
        self.after_save = functools.partial(self.Hook, hook_name='after_save')
        self.after_update = functools.partial(self.Hook, hook_name='after_update')
        self.before_delete = functools.partial(self.Hook, hook_name='before_delete')
        self.after_delete = functools.partial(self.Hook, hook_name='after_delete')
        self._app.add_url_rule('/health', 'health', self.health)
        self._app.add_url_rule('/console/config', 'config', self.config, methods=['GET','POST'])
        self._app.add_url_rule('/console/functionNames', 'functionNames', self.functionNames, methods=['GET','POST'])
        self._app.add_url_rule('/console/jobNames', 'jobNames', self.jobNames, methods=['GET','POST'])
        self._app.add_url_rule('/console/threadStats', 'threadStats', self.threadStats, methods=['GET','POST'])

    def health(self):
        return "ok"

    def jobNames(self):
        return json.dumps(self._job_map.keys())

    def functionNames(self):
        return json.dumps(self._function_map.keys())

    def config(self):
        if os.path.exists("/home/leap/global.json"):
            return open("/home/leap/global.json").read()
        return "{}"

    def threadStats(self):
        return json.dumps({"queueSize":60,"rejectCount":0})

    def Function(self, func):
        def _deco():
            ML.by_hook(False)
            return func(request)

        self._function_map[func.__name__] = _deco
        self._app.add_url_rule(
            '/function/{}'.format(func.__name__),
            'function.{}'.format(func.__name__),
            _deco,
            methods=['POST']
        )
        return _deco

    def Job(self, func):
        def _deco():
            ML.by_hook(False)
            return func(request)

        self._job_map[func.__name__] = _deco
        self._app.add_url_rule(
            '/job/{}'.format(func.__name__),
            'job.{}'.format(func.__name__),
            _deco,
            methods=['POST']
        )
        return _deco

    def Hook(self, class_name, hook_name):
        hook_name = '{}_{}'.format(hook_name, class_name)
        def _deco(func):
            def __deco():
                ML.by_hook(True)
                method = request.json['method']
                params = request.json['params']
                res = self._handel_hook(class_name, method, params)
                ML.by_hook(False)
                return res

            self._hook_map[hook_name] = func

            if class_name not in self._hook_classes:
                self._app.add_url_rule(
                    '/entityManager/{}'.format(class_name),
                    'entityManager.{}'.format(class_name),
                    __deco,
                    methods=['POST']
                )
            self._hook_classes.add(class_name)
            return __deco
        return _deco

    def _handel_hook(self, class_name, method, params):

        if method == 'create':
            params['principal'] = get_principal()
            obj = ML.Object.create(class_name,**params)

            if 'before_save_{}'.format(class_name) in self._hook_map:
                hook_res = self._hook_map['before_save_{}'.format(class_name)](obj)
                if isinstance(hook_res, Response):return hook_res

            obj.save()
            res = Response(json.dumps({
                "createdAt":"{}.{}Z".format(obj.created_at.strftime('%Y-%m-%dT%H:%M:%S'),obj.created_at.microsecond/1000),
                "objectId":obj.id
                }),
                mimetype='application/json'
            )

            if 'after_save_{}'.format(class_name) in self._hook_map:
                hook_res = self._hook_map['after_save_{}'.format(class_name)](obj)
                if isinstance(hook_res, Response):return hook_res


        elif method == 'update':
            attrs = {}
            attrs['objectId'] = params['objectId']
            attrs.update(params['update'])
            attrs['principal'] = get_principal()
            obj = ML.Object.create(class_name,**attrs)
            obj.save()
            res = Response(json.dumps({
                "updatedAt":"{}.{}Z".format(obj.updated_at.strftime('%Y-%m-%dT%H:%M:%S'),obj.updated_at.microsecond/1000),
                "number":1
                }),
                mimetype='application/json'
            )

            if 'after_update_{}'.format(class_name) in self._hook_map:
                hook_res = self._hook_map['after_update_{}'.format(class_name)](obj)
                if isinstance(hook_res, Response):return hook_res

        elif method == 'delete':
            params['principal'] = get_principal()
            obj = ML.Object.create(class_name,**params)
            if 'before_delete_{}'.format(class_name) in self._hook_map:
                hook_res = self._hook_map['before_delete_{}'.format(class_name)](obj)
                if isinstance(hook_res, Response): return hook_res

            obj.destroy()
            res = Response(json.dumps({
                "number":1
                }),
                mimetype='application/json'
            )

            if 'after_delete_{}'.format(class_name) in self._hook_map:
                hook_res = self._hook_map['after_delete_{}'.format(class_name)](obj)
                if isinstance(hook_res, Response): return hook_res
        else:
            raise ML.MaxLeapError(501, "Unknown method [{}]".format(method))
        if isinstance(res,Response):
            return res
        else:
            raise ML.MaxLeapError(510, "hook didn't return an Response Object")

    def run(self):
        if ML.DEBUG:
            self._app.run(host=self._host, port=self._port)
        else:
            raise Runtime('Not Debug Model')

    def callFunction(self, name, *args, **kwargs):
        client = self._app.test_client()
        path = '/function/{0}'.format(name)
        headers = {"Content-Type":"application/json"}
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers
        return client.post(path, *args, **kwargs)

    def callJob(self, name, *args, **kwargs):
        client = self._app.test_client()
        path = '/job/{0}'.format(name)
        headers = {"Content-Type":"application/json"}
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers
        return client.post(path, *args, **kwargs)

Server = flask_base_server(app, Host, Port)



# coding: utf-8

import copy
import json
import gzip
from cStringIO import StringIO
from datetime import datetime

import arrow
import iso8601
from dateutil import tz

import ML
from ML import operation

__author__ = 'czhou <czhou@ilegendsoft.com>'


def get_dumpable_types():
    return (
        ML.ACL,
        ML.GeoPoint,
        ML.Relation,
        operation.BaseOp,
    )


def encode(value, disallow_objects=False):
    if isinstance(value, datetime):
        tzinfo = value.tzinfo
        if tzinfo is None:
            tzinfo = tz.tzlocal()
        return {
            '__type': 'Date',
            'iso': arrow.get(value, tzinfo).to('utc').format('YYYY-MM-DDTHH:mm:ss.SSS') + 'Z',
        }

    if isinstance(value, ML.Object):
        if disallow_objects:
            raise ValueError('ML.Object not allowed')
        return value._to_pointer()

    if isinstance(value, ML.File):
        if not value.url and not value.id:
            raise ValueError('Tried to save an object containing an unsaved file.')
        return {
            '__type': 'File',
            'id': value.id,
            'name': value.name,
            'url': value.url,
        }


    if isinstance(value, get_dumpable_types()):
        return value.dump()

    if isinstance(value, (tuple, list)):
        return [encode(x, disallow_objects) for x in value]

    if isinstance(value, dict):
        return  dict([(k, encode(v, disallow_objects)) for k, v in value.iteritems()])

    return value





def decode(key, value):
    if isinstance(value, get_dumpable_types()):
        return value

    if isinstance(value, (tuple, list)):
        return [decode(key, x) for x in value]

    if not isinstance(value, dict):
        return value

    if '__type' not in value:
        return dict([(k, decode(k, v)) for k, v in value.iteritems()])

    _type = value['__type']

    if _type == 'Pointer':
        value = copy.deepcopy(value)
        class_name = value['className']
        pointer = ML.Object.create(class_name)
        if 'createdAt' in value:
            value.pop('__type')
            value.pop('className')
            pointer._finish_fetch(value, True)
        else:
            pointer._finish_fetch({'objectId': value['objectId']}, False)
        return pointer

    if _type == 'Object':
        value = copy.deepcopy(value)
        class_name = value['className']
        value.pop('__type')
        value.pop('class_name')
        obj = ML.Object.create(class_name)
        obj._finish_fetch(value, True)
        return obj

    if _type == 'Date':
        return arrow.get(iso8601.parse_date(value['iso'])).to('local').datetime

    if _type == 'GeoPoint':
        return ML.GeoPoint(latitude=value['latitude'], longitude=value['longitude'])

    if key == 'ACL':
        if isinstance(value, ML.ACL):
            return value
        return ML.ACL(value)

    if _type == 'Relation':
        relation = ML.Relation(None, key)
        relation.target_class_name = value['className']
        return relation

    if _type == 'File':
        f = ML.File(value['name'])
        meta_data = value.get('metaData')
        if meta_data:
            f._metadata = meta_data
        f._url = value['url']
        f.id = ''
        return f


def traverse_object(obj, callback, seen=None):
    seen = seen or set()

    if isinstance(obj, ML.Object):
        if obj in seen:
            return
        seen.add(obj)
        traverse_object(obj.attributes, callback, seen)
        return callback(obj)

    if isinstance(obj, (ML.Relation, ML.File)):
        return callback(obj)

    if isinstance(obj, (list, tuple)):
        for idx, child in enumerate(obj):
            new_child = traverse_object(child, callback, seen)
            if new_child:
                obj[idx] = new_child
        return callback(obj)

    if isinstance(obj, dict):
        for key, child in obj.iteritems():
            new_child = traverse_object(child, callback, seen)
            if new_child:
                obj[key] = new_child
        return callback(obj)

    return callback(obj)

def response_to_json(response):
    """
    hack for requests in python 2.6
    """
    if isinstance(response, ML.Response):
        return json.loads(response.data)

    content = response.content
    # hack for requests in python 2.6
    if 'application/json' in response.headers.get('Content-Type',''):
        if content[:2] == '\x1f\x8b':
            f = StringIO(content)
            g = gzip.GzipFile(fileobj=f)
            content = g.read()
            g.close()
            f.close()
    return json.loads(content)



# coding: utf-8

import json
import copy

import requests
requests.packages.urllib3.disable_warnings()

import ML
from ML import utils

__author__ = 'czhou <czhou@ilegendsoft.com>'

APP_ID = None
BY_HOOK = None

US_BASE_URL = 'https://api.leap.as'
CN_BASE_URL = 'https://api.leapcloud.cn'
UAT_BASE_URL = 'http://apiuat.leap.as'
DEV_BASE_URL = 'http://apidev.leap.as'

SERVER_VERSION = '2.0'
SDK_VERSION = '1.0.0'
BASE_URL = US_BASE_URL + '/' + SERVER_VERSION
TIMEOUT_SECONDS = 15

MASTER_USER_PRINCIPAL = None

def get_master_principal():
    return MASTER_USER_PRINCIPAL

class UserPrincipal(object):
    """docstring for UserPrincipal"""
    def __init__(self, app_id, client_key=None, master_key=None, session_token=None):
        super(UserPrincipal, self).__init__()
        self.app_id = app_id
        self.client_key = client_key
        self.master_key = master_key
        self.session_token = session_token
        self.headers = {
            'Content-Type':'application/json;charset=utf-8',
            'User-Agent':'MaxLeap Code Python-{0}SDK'.format(ML.__version__)
        }
        self.validate()

    def validate(self):
        if not any([self.client_key, self.master_key, self.session_token]):
            raise RuntimeError('client_key or master_key or session_token must be specified')
        if not self.app_id:
            raise RuntimeError('app_id must be specified')

    def gen_headers(self):
        self.headers['X-LAS-AppId'] = self.app_id
        if self.master_key:
            self.headers['X-LAS-MasterKey'] = self.master_key
        elif self.client_key:
            self.headers['X-LAS-APIKey'] = self.client_key
        elif self.session_token:
            self.headers['X-LAS-Session-Token'] = self.session_token
        return self.headers

    def injection(self,headers):
        self.headers.update(headers)


def change_region(region='us'):
    global BASE_URL
    if region == 'us':
        BASE_URL = US_BASE_URL + '/' + SERVER_VERSION
    if region == 'cn':
        BASE_URL = CN_BASE_URL + '/' + SERVER_VERSION
    if region == 'uat':
        BASE_URL = UAT_BASE_URL + '/' + SERVER_VERSION
    if region == 'dev':
        BASE_URL = DEV_BASE_URL + '/' + SERVER_VERSION

def by_hook(flag):
    """
    :type flag: bool
    :param flag: 当请求从cloudCode发起的时候，标志为True
    """
    global BY_HOOK
    BY_HOOK = flag


def init(app_id, client_key=None, master_key=None, region='us'):
    """初始化 MaxLeap 的 AppId / REST API Key / MasterKey

    :type app_id: basestring
    :param app_id: 应用的 Application ID
    :type client_key: None or basestring
    :param client_key: 应用的 REST API Key
    :type master_key: None or basestring
    :param master_key: 应用的 Master Key
    """
    if (not client_key) and (not master_key):
        raise RuntimeError('client_key or master_key must be specified')
    global MASTER_USER_PRINCIPAL, APP_ID
    APP_ID = app_id
    MASTER_USER_PRINCIPAL = UserPrincipal(app_id, client_key=client_key, master_key=master_key)
    change_region(region)


def need_init(func):
    def new_func(*args, **kwargs):
        if MASTER_USER_PRINCIPAL is None:
            raise RuntimeError('MaxLeap SDK must be initialized')
        if not kwargs.get('principal'):
            kwargs['principal'] = ML.get_principal() or copy.deepcopy(MASTER_USER_PRINCIPAL)
        if BY_HOOK:
            kwargs['principal'].injection({'X-ZCloud-Request-From-Cloudcode':'true'})
        return func(*args, **kwargs)
    return new_func


def check_error(func):
    def new_func(*args, **kwargs):
        response = func(*args, **kwargs)
        if response.headers.get('Content-Type') == 'text/html':
            raise ML.MaxLeapError(-1, 'Bad Request')
        content = utils.response_to_json(response)
        if 'errorCode' in content:
            raise ML.MaxLeapError(content.get('errorCode', 1), content.get('errorMessage', 'Unknown Error'))
        return response
    return new_func

def handler_hook(method):
    def _deco(func):
        def new_func(*args, **kwargs):
            global BY_HOOK
            if ML.PRO or BY_HOOK:
                return func(*args, **kwargs)
            else:
                path = args[0]
                url_parts = path.split('/')
                obj_id = None
                class_name = None
                if len(url_parts) == 3:
                    class_name = url_parts[2]
                elif len(url_parts) == 4:
                    class_name = url_parts[2]
                    obj_id = url_parts[3]
                else:
                    return func(*args, **kwargs)

                if class_name not in ML.Server._hook_classes:
                    return func(*args, **kwargs)

                if method == "create":
                    BY_HOOK = True
                    res = ML.Server._handel_hook(class_name, method, args[1])
                    BY_HOOK = False
                    return res

                if method == "update":
                    BY_HOOK = True
                    res = ML.Server._handel_hook(class_name, method, {"update": args[1], "objectId": obj_id})
                    BY_HOOK = False
                    return res

                if method == "delete":
                    BY_HOOK = True
                    res = ML.Server._handel_hook(class_name, method, {"objectId": obj_id})
                    BY_HOOK = False
                    return res

                return func(*args, **kwargs)
        return new_func
    return _deco


@need_init
@check_error
def get(url, params, principal=None):
    for k, v in params.iteritems():
        if isinstance(v, dict):
            params[k] = json.dumps(v)
    response = requests.get(BASE_URL + url, headers=principal.gen_headers(), params=params, timeout=TIMEOUT_SECONDS, verify=False)
    return response

@need_init
@check_error
@handler_hook('create')
def post(url, params, principal=None):
    response = requests.post(BASE_URL + url, headers=principal.gen_headers(), data=json.dumps(params), timeout=TIMEOUT_SECONDS, verify=False)
    return response

@need_init
@check_error
@handler_hook('update')
def put(url, params, principal=None):
    response = requests.put(BASE_URL + url, headers=principal.gen_headers(), data=json.dumps(params), timeout=TIMEOUT_SECONDS, verify=False)
    return response

@need_init
@check_error
@handler_hook('delete')
def delete(url, params=None, principal=None):
    response = requests.delete(BASE_URL + url, headers=principal.gen_headers(), data=json.dumps(params), timeout=TIMEOUT_SECONDS, verify=False)
    return response



# coding: utf-8

import os
import re
import base64
import cStringIO
import StringIO

import ML
from ML import client
from ML import utils
from ML.mime_type import mime_types


__author__ = 'czhou <czhou@ilegendsoft.com>'


class File(object):
    def __init__(self, name, data=None, type_=None):
        self._name = name
        self.id = None
        self._url = None
        self._acl = None
        self.current_user = None  # TODO
        self._metadata = {
            'owner': 'unknown'
        }
        if self.current_user and self.current_user is not None:
            self._metadata['owner'] = self.current_user.id

        pattern = re.compile('\.([^.]*)$')
        extension = pattern.findall(name)
        if extension:
            extension = extension[0].lower()
        else:
            extension = None

        if type_:
            self._type = type_
        else:
            self._type = mime_types.get(extension, 'text/plain')

        if data is None:
            self._source = None
        elif isinstance(data, cStringIO.OutputType):
            self._source = StringIO.StringIO(data.getvalue())
        elif isinstance(data, StringIO.StringIO):
            self._source = data
        elif isinstance(data, file):
            data.seek(0, os.SEEK_SET)
            self._source = StringIO.StringIO(data.read())
        elif isinstance(data, buffer):
            self._source = StringIO.StringIO(data)
        else:
            raise TypeError('data must be a StringIO / buffer / file instance')

        if self._source:
            self._source.seek(0, os.SEEK_END)
            self._metadata['size'] = self._source.tell()
            self._source.seek(0, os.SEEK_SET)

    @classmethod
    def create_with_url(cls, name, url, meta_data=None, type_=None):
        f = File(name, None, type_)
        if meta_data:
            f._metadata.update(meta_data)

        f._url = url
        f._metadata['__source'] = 'external'
        return f

    @classmethod
    def create_without_data(cls, object_id):
        f = File('')
        f.id = object_id
        return f

    def get_acl(self):
        return self._acl

    def set_acl(self, acl):
        if not isinstance(acl, ML.ACL):
            raise TypeError('acl must be a ML.ACL instance')
        self._acl = acl

    @property
    def name(self):
        return self._name

    @property
    def url(self):
        return self._url

    @property
    def size(self):
        return self._metadata['size']

    @property
    def owner_id(self):
        return self._metadata['owner']

    @property
    def metadata(self):
        return self._metadata

    def get_thumbnail_url(self, width, height, quality=100, scale_to_fit=True, fmt='png'):
        if not self._url:
            raise ValueError('invalid url')

        if width < 0 or height < 0:
            raise ValueError('invalid height or width params')

        if quality > 100 or quality <= 0:
            raise ValueError('quality must between 0 and 100')

        mode = 2 if scale_to_fit else 1

        return self.url + '?imageView/{0}/w/{1}/h/{2}/q/{3}/format/{4}'.format(mode, width, height, quality, fmt)

    def destroy(self):
        if not self.id:
            return False
        response = client.delete('/files/{0}'.format(self.id))
        content = utils.response_to_json(response)
        return response

    def save(self):
        if self._source:
            output = cStringIO.StringIO()
            self._source.seek(0)
            base64.encode(self._source, output)
            self._source.seek(0)
            data = {
                'base64': output.getvalue(),
                '_ContentType': self._type,
                'mime_type': self._type,
                'metaData': self._metadata,
            }
        elif self._url and self.metadata['__source'] == 'external':
            data = {
                'name': self._name,
                'ACL': self._acl,
                'metaData': self._metadata,
                'mime_type': self._type,
                'url': self._url,
            }
        else:
            raise ValueError

        response = client.put('/files/{}'.format(self._name), data)
        content = utils.response_to_json(response)

        self._name = content['name'] #平台返回的 name和File文件的name不一致
        self._url = content['url']
        self.id = content['uid']
        if 'size' in content:
            self._metadata['size'] = content['size']

        return self



# coding: utf-8

__author__ = 'czhou <czhou@ilegendsoft.com>'

mime_types = {
    "ai": "application/postscript",
    "aif": "audio/x-aiff",
    "aifc": "audio/x-aiff",
    "aiff": "audio/x-aiff",
    "asc": "text/plain",
    "atom": "application/atom+xml",
    "au": "audio/basic",
    "avi": "video/x-msvideo",
    "bcpio": "application/x-bcpio",
    "bin": "application/octet-stream",
    "bmp": "image/bmp",
    "cdf": "application/x-netcdf",
    "cgm": "image/cgm",
    "class": "application/octet-stream",
    "cpio": "application/x-cpio",
    "cpt": "application/mac-compactpro",
    "csh": "application/x-csh",
    "css": "text/css",
    "dcr": "application/x-director",
    "dif": "video/x-dv",
    "dir": "application/x-director",
    "djv": "image/vnd.djvu",
    "djvu": "image/vnd.djvu",
    "dll": "application/octet-stream",
    "dmg": "application/octet-stream",
    "dms": "application/octet-stream",
    "doc": "application/msword",
    "docx": "application/vnd.openxmlformats-officedocument.wordprocessingml." +
           "document",
    "dotx": "application/vnd.openxmlformats-officedocument.wordprocessingml." +
           "template",
    "docm": "application/vnd.ms-word.document.macroEnabled.12",
    "dotm": "application/vnd.ms-word.template.macroEnabled.12",
    "dtd": "application/xml-dtd",
    "dv": "video/x-dv",
    "dvi": "application/x-dvi",
    "dxr": "application/x-director",
    "eps": "application/postscript",
    "etx": "text/x-setext",
    "exe": "application/octet-stream",
    "ez": "application/andrew-inset",
    "gif": "image/gif",
    "gram": "application/srgs",
    "grxml": "application/srgs+xml",
    "gtar": "application/x-gtar",
    "hdf": "application/x-hdf",
    "hqx": "application/mac-binhex40",
    "htm": "text/html",
    "html": "text/html",
    "ice": "x-conference/x-cooltalk",
    "ico": "image/x-icon",
    "ics": "text/calendar",
    "ief": "image/ief",
    "ifb": "text/calendar",
    "iges": "model/iges",
    "igs": "model/iges",
    "jnlp": "application/x-java-jnlp-file",
    "jp2": "image/jp2",
    "jpe": "image/jpeg",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "js": "application/x-javascript",
    "kar": "audio/midi",
    "latex": "application/x-latex",
    "lha": "application/octet-stream",
    "lzh": "application/octet-stream",
    "m3u": "audio/x-mpegurl",
    "m4a": "audio/mp4a-latm",
    "m4b": "audio/mp4a-latm",
    "m4p": "audio/mp4a-latm",
    "m4u": "video/vnd.mpegurl",
    "m4v": "video/x-m4v",
    "mac": "image/x-macpaint",
    "man": "application/x-troff-man",
    "mathml": "application/mathml+xml",
    "me": "application/x-troff-me",
    "mesh": "model/mesh",
    "mid": "audio/midi",
    "midi": "audio/midi",
    "mif": "application/vnd.mif",
    "mov": "video/quicktime",
    "movie": "video/x-sgi-movie",
    "mp2": "audio/mpeg",
    "mp3": "audio/mpeg",
    "mp4": "video/mp4",
    "mpe": "video/mpeg",
    "mpeg": "video/mpeg",
    "mpg": "video/mpeg",
    "mpga": "audio/mpeg",
    "ms": "application/x-troff-ms",
    "msh": "model/mesh",
    "mxu": "video/vnd.mpegurl",
    "nc": "application/x-netcdf",
    "oda": "application/oda",
    "ogg": "application/ogg",
    "pbm": "image/x-portable-bitmap",
    "pct": "image/pict",
    "pdb": "chemical/x-pdb",
    "pdf": "application/pdf",
    "pgm": "image/x-portable-graymap",
    "pgn": "application/x-chess-pgn",
    "pic": "image/pict",
    "pict": "image/pict",
    "png": "image/png",
    "pnm": "image/x-portable-anymap",
    "pnt": "image/x-macpaint",
    "pntg": "image/x-macpaint",
    "ppm": "image/x-portable-pixmap",
    "ppt": "application/vnd.ms-powerpoint",
    "pptx": "application/vnd.openxmlformats-officedocument.presentationml." +
           "presentation",
    "potx": "application/vnd.openxmlformats-officedocument.presentationml." +
           "template",
    "ppsx": "application/vnd.openxmlformats-officedocument.presentationml." +
           "slideshow",
    "ppam": "application/vnd.ms-powerpoint.addin.macroEnabled.12",
    "pptm": "application/vnd.ms-powerpoint.presentation.macroEnabled.12",
    "potm": "application/vnd.ms-powerpoint.template.macroEnabled.12",
    "ppsm": "application/vnd.ms-powerpoint.slideshow.macroEnabled.12",
    "ps": "application/postscript",
    "qt": "video/quicktime",
    "qti": "image/x-quicktime",
    "qtif": "image/x-quicktime",
    "ra": "audio/x-pn-realaudio",
    "ram": "audio/x-pn-realaudio",
    "ras": "image/x-cmu-raster",
    "rdf": "application/rdf+xml",
    "rgb": "image/x-rgb",
    "rm": "application/vnd.rn-realmedia",
    "roff": "application/x-troff",
    "rtf": "text/rtf",
    "rtx": "text/richtext",
    "sgm": "text/sgml",
    "sgml": "text/sgml",
    "sh": "application/x-sh",
    "shar": "application/x-shar",
    "silo": "model/mesh",
    "sit": "application/x-stuffit",
    "skd": "application/x-koan",
    "skm": "application/x-koan",
    "skp": "application/x-koan",
    "skt": "application/x-koan",
    "smi": "application/smil",
    "smil": "application/smil",
    "snd": "audio/basic",
    "so": "application/octet-stream",
    "spl": "application/x-futuresplash",
    "src": "application/x-wais-source",
    "sv4cpio": "application/x-sv4cpio",
    "sv4crc": "application/x-sv4crc",
    "svg": "image/svg+xml",
    "swf": "application/x-shockwave-flash",
    "t": "application/x-troff",
    "tar": "application/x-tar",
    "tcl": "application/x-tcl",
    "tex": "application/x-tex",
    "texi": "application/x-texinfo",
    "texinfo": "application/x-texinfo",
    "tif": "image/tiff",
    "tiff": "image/tiff",
    "tr": "application/x-troff",
    "tsv": "text/tab-separated-values",
    "txt": "text/plain",
    "ustar": "application/x-ustar",
    "vcd": "application/x-cdlink",
    "vrml": "model/vrml",
    "vxml": "application/voicexml+xml",
    "wav": "audio/x-wav",
    "wbmp": "image/vnd.wap.wbmp",
    "wbmxl": "application/vnd.wap.wbxml",
    "wml": "text/vnd.wap.wml",
    "wmlc": "application/vnd.wap.wmlc",
    "wmls": "text/vnd.wap.wmlscript",
    "wmlsc": "application/vnd.wap.wmlscriptc",
    "wrl": "model/vrml",
    "xbm": "image/x-xbitmap",
    "xht": "application/xhtml+xml",
    "xhtml": "application/xhtml+xml",
    "xls": "application/vnd.ms-excel",
    "xml": "application/xml",
    "xpm": "image/x-xpixmap",
    "xsl": "application/xml",
    "xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "xltx": "application/vnd.openxmlformats-officedocument.spreadsheetml." +
          "template",
    "xlsm": "application/vnd.ms-excel.sheet.macroEnabled.12",
    "xltm": "application/vnd.ms-excel.template.macroEnabled.12",
    "xlam": "application/vnd.ms-excel.addin.macroEnabled.12",
    "xlsb": "application/vnd.ms-excel.sheet.binary.macroEnabled.12",
    "xslt": "application/xslt+xml",
    "xul": "application/vnd.mozilla.xul+xml",
    "xwd": "image/x-xwindowdump",
    "xyz": "chemical/x-xyz",
    "zip": "application/zip"
}



# coding: utf-8

import copy
from datetime import datetime

import iso8601

import ML
from ML import utils
from ML import client
from ML import operation


__author__ = 'czhou <czhou@ilegendsoft.com>'


object_class_map = {}


class ObjectMeta(type):
    def __new__(cls, name, bases, attrs):
        cached_class = object_class_map.get(name)
        if cached_class:
            return cached_class

        super_new = super(ObjectMeta, cls).__new__
        if name == 'User':
            attrs['_class_name'] = '_User'
        elif name == 'Installation':
            attrs['_class_name'] = '_Installation'
        elif name == 'Notification':
            attrs['_class_name'] = '_Notification'
        else:
            attrs['_class_name'] = name
        object_class = super_new(cls, name, bases, attrs)
        object_class_map[name] = object_class
        return object_class


class Object(object):
    __metaclass__ = ObjectMeta

    def __init__(self, principal=None, **attrs):
        """
        创建一个新的 ML.Object

        :param attrs: 对象属性
        :return:
        """
        self.id = None
        self._class_name = self._class_name  # for IDE

        self._server_data = {}
        self._op_set_queue = [{}]
        self.attributes = {}

        self._existed = False

        self.created_at = None
        self.updated_at = None
        self.principal = principal

        for k, v in attrs.iteritems():
            self.set(k, v)

    @classmethod
    def extend(cls, name):
        """
        派生一个新的 ML.Object 子类

        :param name: 子类名称
        :type name: basestring
        :return: 派生的子类
        :rtype: ObjectMeta
        """
        if isinstance(name, unicode):
            name = name.encode('utf-8')
        return type(name, (cls,), {})

    @classmethod
    def create(cls, class_name, **attributes):
        """
        根据参数创建一个 ML.Object 的子类的实例化对象

        :param class_name: 子类名称
        :type class_name: basestring
        :param attributes: 对象属性
        :return: 派生子类的实例
        :rtype: Object
        """
        object_class = cls.extend(class_name)
        return object_class(**attributes)

    @classmethod
    def create_without_data(cls, id_):
        """
        根据 objectId 创建一个 ML.Object，代表一个服务器上已经存在的对象。可以调用 fetch 方法来获取服务器上的数据

        :param id_: 对象的 objectId
        :type id_: basestring
        :return: 没有数据的对象
        :rtype: Object
        """
        if cls is Object:
            raise RuntimeError('can not call create_without_data on ML.Object')
        obj = cls()
        obj.id = id_
        return obj

    def dump(self):
        obj = self._dump()
        obj.pop('__type')
        obj.pop('className')
        return obj

    def _dump(self, seen_objects=None):
        seen_objects = seen_objects or []
        obj = copy.deepcopy(self.attributes)
        for k, v in obj.iteritems():
            obj[k] = utils.encode(v, seen_objects)

        if self.id is not None:
            obj['objectId'] = self.id

        obj['__type'] = 'Object'
        obj['className'] = self.__class__.__name__
        return obj

    def destroy(self):
        """
        从服务器上删除这个对象

        :rtype: response
        """
        if not self.id:
            return False
        client.delete('/classes/{0}/{1}'.format(self._class_name, self.id), principal=self.principal)

    def save(self):
        """
        将对象数据保存至服务器

        :return: None
        :rtype: response
        """
        unsaved_children = []
        unsaved_files = []

        self._find_unsaved_children(self.attributes, unsaved_children, unsaved_files)

        if len(unsaved_children) + len(unsaved_files) > 0:
            self._deep_save(self.attributes)

        self._start_save()

        data = self._dump_save()

        method = 'PUT' if self.id is not None else 'POST'

        if method == 'PUT':
            response = client.put('/classes/{0}/{1}'.format(self._class_name, self.id), data, principal=self.principal)
        else:
            response = client.post('/classes/{0}'.format(self._class_name), data, principal=self.principal)

        self._finish_save(self.parse(utils.response_to_json(response), response.status_code))

    def _deep_save(self, exclude=None):
        unsaved_children = []
        unsaved_files = []
        self._find_unsaved_children(self.attributes, unsaved_children, unsaved_files)

        if exclude:
            unsaved_children = [x for x in unsaved_children if x != exclude]

        for f in unsaved_files:
            f.save()



        dumped_objs = []
        for obj in unsaved_children:
            obj._start_save()
            method = 'POST' if obj.id is None else 'PUT'
            path = '/{0}/classes/{1}'.format(client.SERVER_VERSION, obj._class_name)
            body = obj._dump_save()
            dumped_obj = {
                'method': method,
                'path': path,
                'body': body,
            }
            dumped_objs.append(dumped_obj)

        if dumped_objs:
            response = utils.response_to_json(client.post('/batch', params={'requests': dumped_objs}, principal=self.principal))

            errors = []
            for idx, obj in enumerate(unsaved_children):
                content = response[idx]
                if content.get('errorCode'):
                    errors.append(leap.MaxLeapError(content.get('errorCode'), content.get('errorMessage')))
                    obj._cancel_save()
                else:
                    result = obj.parse(content)
                    obj._finish_save(result)

                if errors:
                    # TODO: how to raise list of errors?
                    raise errors[0]

    @classmethod
    def _find_unsaved_children(cls, obj, children, files):

        def callback(o):
            if isinstance(o, Object):
                if o.is_dirty():
                    children.append(o)
                return

            if isinstance(o, ML.File):
                if o.url is None and o.id is None:
                    files.append(o)
                return

        utils.traverse_object(obj, callback)

    def is_dirty(self, attr=None):
        current_changes = self._op_set_queue[-1]

        if attr is not None:
            return True if attr in current_changes else False

        if self.id is None:
            return True

        if current_changes:
            return True

        return False

    def _to_pointer(self):
        return {
            '__type': 'Pointer',
            'className': self._class_name,
            'objectId': self.id,
        }

    def _merge_magic_field(self, attrs):
        for key in ['id', 'objectId', 'createdAt', 'updatedAt']:
            if attrs.get(key) is None:
                continue
            if key == 'objectId':
                self.id = attrs[key]
            elif key == 'createdAt' or key == 'updatedAt':
                if not isinstance(attrs[key], datetime):
                    dt = iso8601.parse_date(attrs[key])
                else:
                    dt = attrs[key]
                if key == 'createdAt':
                    setattr(self, 'created_at', dt)
                elif key == 'updatedAt':
                    setattr(self, 'updated_at', dt)
            del attrs[key]

    def _start_save(self):
        self._op_set_queue.append({})

    def _cancel_save(self):
        failed_changes = self._op_set_queue.pop(0)
        next_changes = self._op_set_queue[0]
        for key, op in failed_changes.iteritems():
            op1 = failed_changes[key]
            op2 = next_changes[key]
            if op1 and op2:
                next_changes[key] = op2._merge(op1)
            elif op1:
                next_changes[key] = op1

    def validate(self, attrs):
        # if 'ACL' in attrs and not isinstance(attrs['ACL'], ML.ACL):
            # raise TypeError('acl must be a ACL')
        return True

    def get(self, attr):
        """
        获取对象字段的值

        :param attr: 字段名
        :type attr: basestring
        :return: 字段值
        """
        return self.attributes.get(attr)

    def relation(self, attr):
        """
        返回对象上相应字段的 Relation

        :param attr: 字段名
        :type attr: basestring
        :return: Relation
        :rtype: ML.Relation
        """
        value = self.get(attr)
        if value is not None:
            if not isinstance(value, ML.Relation):
                raise TypeError('field %s is not Relation'.format(attr))
            value._ensure_parent_and_key(self, attr)
            return value
        return ML.Relation(self, attr)

    def has(self, attr):
        """
        判断此字段是否有值

        :param attr: 字段名
        :return: 当有值时返回 True， 否则返回 False
        :rtype: bool
        """
        return attr in self.attributes

    def set(self, key_or_attrs, value=None, unset=False):
        """
        在当前对象此字段上赋值

        :param key_or_attrs: 字段名，或者一个包含 字段名 / 值的 dict
        :type key_or_attrs: basestring or dict
        :param value: 字段值
        :param unset:
        :return: 当前对象，供链式调用
        """
        if isinstance(key_or_attrs, dict) and value is None:
            attrs = key_or_attrs
        else:
            key = key_or_attrs
            attrs = {key: utils.decode(key, value)}

        if unset:
            for k in attrs.keys():
                attrs[k] = operation.Unset()

        self.validate(attrs)

        self._merge_magic_field(attrs)

        keys = attrs.keys()
        for k in keys:
            v = attrs[k]

            if not isinstance(v, operation.BaseOp):
                v = operation.Set(v)

            current_changes = self._op_set_queue[-1]
            current_changes[k] = v._merge(current_changes.get(k))
            self._rebuild_estimated_data_for_key(k)

        return self

    def unset(self, attr):
        """
        在对象上移除此字段。

        :param attr: 字段名
        :return: 当前对象
        """
        return self.set(attr, None, unset=True)

    def increment(self, attr, amount=1):
        """
        在对象此字段上自增对应的数值，如果数值没有指定，默认为一。

        :param attr: 字段名
        :param amount: 自增量
        :return: 当前对象
        """
        return self.set(attr, operation.Increment(amount))

    def add(self, attr, item):
        """
        在对象此字段对应的数组末尾添加指定对象。

        :param attr: 字段名
        :param item: 要添加的对象
        :return: 当前对象
        """
        return self.set(attr, operation.Add([item]))

    def add_unique(self, attr, item):
        """
        在对象此字段对应的数组末尾添加指定对象，如果此对象并没有包含在字段中。

        :param attr: 字段名
        :param item: 要添加的对象
        :return: 当前对象
        """
        return self.set(attr, operation.AddUnique([item]))

    def remove(self, attr, item):
        """
        在对象此字段对应的数组中，将指定对象全部移除。

        :param attr: 字段名
        :param item: 要移除的对象
        :return: 当前对象
        """
        return self.set(attr, operation.Remove([item]))

    def op(self, attr):
        return self._op_set_queue[-1][attr]

    def clear(self):
        """
        将当前对象所有字段全部移除。

        :return: 当前对象
        """
        self.set(self.attributes, unset=True)

    def _dump_save(self):
        result = copy.deepcopy(self._op_set_queue[0])
        for k, v in result.iteritems():
            result[k] = v.dump()
        return result

    def fetch(self):
        """
        从服务器获取当前对象所有的值，如果与本地值不同，将会覆盖本地的值。

        :return: 当前对象
        """
        response = client.get('/classes/{0}/{1}'.format(self._class_name, self.id), {}, principal=self.principal)
        result = self.parse(utils.response_to_json(response), response.status_code)
        self._finish_fetch(result, True)

    def parse(self, content, status_code=None):
        self._existed = True
        if status_code == 201:
            self._existed = False

        return content

    def is_new(self):
        """
        判断当前对象是否已经保存至服务器。

        :rtype: bool
        """
        return True if self.id else False

    def is_existed(self):
        return self._existed

    def get_acl(self):
        """
        返回当前对象的 ACL。

        :return: 当前对象的 ACL
        :rtype: ML.ACL
        """
        return self.get('ACL')

    def set_acl(self, acl):
        """
        为当前对象设置 ACL

        :type acl: ML.ACL
        :return: 当前对象
        """

        return self.set('ACL', acl)

    def _finish_save(self, server_data):
        saved_changes = self._op_set_queue[0]
        self._op_set_queue = self._op_set_queue[1:]
        self._apply_op_set(saved_changes, self._server_data)
        self._merge_magic_field(server_data)
        self._rebuild_all_estimated_data()

    def _finish_fetch(self, server_data, has_data):
        self._op_set_queue = [{}]

        self._merge_magic_field(server_data)

        for key, value in server_data.iteritems():
            self._server_data[key] = utils.decode(key, value)

        self._rebuild_all_estimated_data()

        self._op_set_queue = [{}]

        self._has_data = has_data

    def _rebuild_estimated_data_for_key(self, key):
        if self.attributes.get(key):
            del self.attributes[key]

        if key in self._server_data:
            self.attributes[key] = self._server_data[key]

        for op_set in self._op_set_queue:
            o = op_set.get(key)
            if o is None:
                continue
            self.attributes[key] = o._apply(self.attributes.get(key), self, key)
            if self.attributes[key] is operation._UNSET:
                del self.attributes[key]

    def _rebuild_all_estimated_data(self):
        self.attributes = copy.deepcopy(self._server_data)

    def _apply_op_set(self, op_set, target):
        for key, change in op_set.iteritems():
            target[key] = change._apply(target.get(key), self, key)
            if target[key] == operation._UNSET:
                del target[key]



# coding: utf-8

from ML import client
from ML import Object
from ML import utils

__author__ = 'czhou'


class User(Object):
    def __init__(self, **attrs):
        self._is_current_user = False
        self._session_token = None
        super(User, self).__init__(**attrs)

    def _merge_magic_field(self, attrs):
        self._session_token = attrs.get('sessionToken')
        attrs.pop('sessionToken', None)

        return super(User, self)._merge_magic_field(attrs)

    @property
    def is_current(self):
        return self._is_current_user

    def _cleanup_auth_data(self):
        if not self.is_current:
            return
        auth_data = self.get('authData')
        if not auth_data:
            return
        keys = auth_data.keys()
        for key in keys:
            if not auth_data[key]:
                del auth_data[key]

    def _sync_all_auth_data(self):
        auth_data = self.get('authData')
        if not auth_data:
            return
        for key in auth_data.keys():
            self._sync_auth_data(key)

    def _sync_auth_data(self, key):
        if not self.is_current:
            return

    def _handle_save_result(self, make_current=False):
        if make_current:
            self._is_current_user = True
        self._cleanup_auth_data()
        self._sync_all_auth_data()
        self._server_data.pop('password', None)
        self._rebuild_estimated_data_for_key('password')

    def save(self):
        super(User, self).save()
        self._handle_save_result(False)

    def sign_up(self):
        """
        创建一个新用户。新创建的 User 对象，应该使用此方法来将数据保存至服务器，而不是使用 save 方法。
        用户对象上必须包含 username 和 password 两个字段
        """
        username = self.get('username')
        if not username:
            raise TypeError('invalid username: {0}'.format(username))

        password = self.get('password')
        if not password:
            raise TypeError('invalid password')

        self.save()

    def login(self, username=None, password=None):
        """
        登陆用户。如果用户名和密码正确，服务器会返回用户的 sessionToken 。
        """
        if username:
            self.set('username', username)
        if password:
            self.set('password', password)
        response = client.get('/login', params=self.dump())
        content = utils.response_to_json(response)
        server_data = self.parse(content, response.status_code)
        self._finish_fetch(server_data, False)
        self._handle_save_result(True)
        if 'smsCode' not in server_data:
            self.attributes.pop('smsCode', None)



# coding: utf-8

__author__ = 'czhou <czhou@ilegendsoft.com>'


class MaxLeapError(Exception):
    def __init__(self, code, error):
        self.code = code
        self.error = error

    def __str__(self):
        error = self.error if isinstance(self.error, str) else self.error.encode('utf-8', 'ignore')
        return '[{0}] {1}'.format(self.code, error)



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='maxleap-sdk',
    version='0.1.11',
    description='MaxLeap Python SDK',

    url='https://leap.as/',

    author='czhou',
    author_email='czhou@ilegendsoft.com',

    license='LGPL',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='MaxLeap SDK',

    packages=['ML'],

    package_dir={'ML':'ML'},

    package_data={'ML':['*.py']},

    test_suite='nose.collector',

    install_requires=[
        'requests',
        'iso8601',
        'arrow',
        'flask',
    ],

    extras_require={
        'dev': ['sphinx'],
        'test': ['nose', 'coverage'],
    },
)



# coding: utf-8

import json

import ML
from ML import client
from ML import utils
from ML.object_ import Object
from ML.errors import MaxLeapError

__author__ = 'czhou <czhou@ilegendsoft.com>'


class Query(object):
    def __init__(self, query_class, principal=None):
        """

        :param query_class: 要查询的 class 名称或者对象
        :type query_class: basestring or ML.ObjectMeta
        """
        if isinstance(query_class, basestring):
            query_class = Object.extend(query_class)
        self._query_class = query_class

        self._where = {}
        self._include = []
        self._limit = -1
        self._skip = 0
        self._extra = {}
        self._order = []
        self._select = []
        self.principal = principal

    @classmethod
    def or_(cls, *queries):
        """
        根据传入的 Query 对象，构造一个新的 OR 查询。

        :param queries: 需要构造的子查询列表
        :rtype: Query
        """
        if len(queries) < 2:
            raise ValueError('or_ need two queries at least')
        if not reduce(lambda l, r: l['className'] == r['className'], queries):
            raise TypeError('All queries must be for the same class')
        query = Query(queries[0]['className'])
        query._or_query(queries)
        return query

    @classmethod
    def and_(cls, *queries):
        """
        根据传入的 Query 对象，构造一个新的 AND 查询。

        :param queries: 需要构造的子查询列表
        :rtype: Query
        """
        if len(queries) < 2:
            raise ValueError('or_ need two queries at least')
        if not reduce(lambda l, r: l['className'] == r['className'], queries):
            raise TypeError('All queries must be for the same class')
        query = Query(queries[0]['className'])
        query._and_query(queries)
        return query


    def dump(self):
        """
        :return: 当前对象的序列化结果
        :rtype: dict
        """
        params = {
            'where': self._where,
        }
        if self._include:
            params['include'] = ','.join(self._include)
        if self._select:
            params['keys'] = ','.join(self._select)
        if self._limit >= 0:
            params['limit'] = self._limit
        if self._skip > 0:
            params['skip'] = self._skip
        if self._order:
            params['order'] = self._order
        params.update(self._extra)
        return params

    def _new_object(self):
        return self._query_class(principal=self.principal)

    def _process_result(self, obj):
        return obj

    def first(self):
        """
        根据查询获取最多一个对象。

        :return: 查询结果
        :rtype: Object
        :raise: MaxLeapError
        """
        params = self.dump()
        params['limit'] = 1
        content = utils.response_to_json(client.get('/classes/{0}'.format(self._query_class._class_name), params, principal=self.principal))
        results = content['results']
        if not results:
            raise MaxLeapError(101, 'Object not found')
        obj = self._new_object()
        obj._finish_fetch(self._process_result(results[0]), True)
        return obj

    def get(self, object_id):
        """
        根据 objectId 查询。

        :param object_id: 要查询对象的 objectId
        :return: 查询结果
        :rtype: Object
        """
        self.equal_to('objectId', object_id)
        return self.first()

    def find(self):
        """
        根据查询条件，获取包含所有满足条件的对象。

        :rtype: list
        """
        content = utils.response_to_json(client.get('/classes/{0}'.format(self._query_class._class_name), self.dump(), principal=self.principal))

        objs = []
        for result in content['results']:
            obj = self._new_object()
            obj._finish_fetch(self._process_result(result), True)
            objs.append(obj)

        return objs

    def destroy_all(self):
        """
        在服务器上删除所有满足查询条件的对象。

        :raise: MaxLeapError
        """
        result = client.delete('/classes/{0}'.format(self._query_class._class_name), self.dump(), principal=self.principal)
        return result

    def count(self):
        """
        返回满足查询条件的对象的数量。

        :rtype: int
        """
        params = self.dump()
        params['limit'] = 0
        params['count'] = 1
        response = client.get('/classes/{0}'.format(self._query_class._class_name), params, principal=self.principal)
        return utils.response_to_json(response)['count']

    def skip(self, n):
        """
        查询条件中跳过指定个数的对象，在做分页时很有帮助。

        :param n: 需要跳过对象的个数
        :rtype: Query
        """
        self._skip = n
        return self

    def limit(self, n):
        """
        设置查询返回结果的数量。如果不设置，默认为 100。最大返回数量为 1000，如果超过这个数量，需要使用多次查询来获取结果。

        :param n: 限制结果的数量
        :rtype: Query
        """
        if n > 1000:
            raise ValueError('limit only accept number less than or equal to 1000')
        self._limit = n
        return self

    def equal_to(self, key, value):
        """
        增加查询条件，查询字段的值必须为指定值。

        :param key: 查询条件的字段名
        :param value: 查询条件的值
        :rtype: Query
        """
        self._where[key] = utils.encode(value)
        return self

    def _add_condition(self, key, condition, value):
        if not self._where.get(key):
            self._where[key] = {}
        self._where[key][condition] = utils.encode(value)
        return self

    def not_equal_to(self, key, value):
        """
        增加查询条件，限制查询结果指定字段的值与查询值不同

        :param key: 查询条件字段名
        :param value: 查询条件值
        :rtype: Query
        """
        self._add_condition(key, '$ne', value)
        return self

    def less_than(self, key, value):
        """
        增加查询条件，限制查询结果指定字段的值小于查询值

        :param key: 查询条件字段名
        :param value: 查询条件值
        :rtype: Query
        """
        self._add_condition(key, '$lt', value)
        return self

    def greater_than(self, key, value):
        """
        增加查询条件，限制查询结果指定字段的值大于查询值

        :param key: 查询条件字段名
        :param value: 查询条件值
        :rtype: Query
        """
        self._add_condition(key, '$gt', value)
        return self

    def less_than_or_equal_to(self, key, value):
        """
        增加查询条件，限制查询结果指定字段的值小于等于查询值

        :param key: 查询条件字段名
        :param value: 查询条件值
        :rtype: Query
        """
        self._add_condition(key, '$lte', value)
        return self

    def greater_than_or_equal_to(self, key, value):
        """
        增加查询条件，限制查询结果指定字段的值大于等于查询值

        :param key: 查询条件字段名
        :param value: 查询条件值名
        :rtype: Query
        """
        self._add_condition(key, '$gte', value)
        return self

    def contained_in(self, key, values):
        """
        增加查询条件，限制查询结果指定字段的值在查询值列表中

        :param key: 查询条件字段名
        :param values: 查询条件值
        :type values: list or tuple
        :rtype: Query
        """
        self._add_condition(key, '$in', values)
        return self

    def not_contained_in(self, key, values):
        """
        增加查询条件，限制查询结果指定字段的值不在查询值列表中

        :param key: 查询条件字段名
        :param values: 查询条件值
        :type values: list or tuple
        :rtype: Query
        """
        self._add_condition(key, '$nin', values)
        return self

    def contains_all(self, key, values):
        """
        增加查询条件，限制查询结果指定字段的值全部包含与查询值列表中

        :param key: 查询条件字段名
        :param values: 查询条件值
        :type values: list or tuple
        :rtype: Query
        """
        self._add_condition(key, '$all', values)
        return self

    def exists(self, key):
        """
        增加查询条件，限制查询结果对象包含指定字段

        :param key: 查询条件字段名
        :rtype: Query
        """
        self._add_condition(key, '$exists', True)
        return self

    def does_not_exists(self, key):
        """
        增加查询条件，限制查询结果对象不包含指定字段

        :param key: 查询条件字段名
        :rtype: Query
        """
        self._add_condition(key, '$exists', False)
        return self

    def matched(self, key, regex, ignore_case=False, multi_line=False):
        """
        增加查询条件，限制查询结果对象指定字段满足指定的正则表达式。

        :param key: 查询条件字段名
        :param regex: 查询正则表达式
        :param ignore_case: 查询是否忽略大小写，默认不忽略
        :param multi_line: 查询是否陪陪多行，默认不匹配
        :rtype: Query
        """
        if not isinstance(regex, basestring):
            raise TypeError('matched only accept str or unicode')
        self._add_condition(key, '$regex', regex)
        modifiers = ''
        if ignore_case:
            modifiers += 'i'
        if multi_line:
            modifiers += 'm'
        if modifiers:
            self._add_condition(key, '$options', modifiers)
        return self

    def matches_query(self, key, query):
        """
        增加查询条件，限制查询结果对象指定字段的值，与另外一个查询对象的返回结果相同。

        :param key: 查询条件字段名
        :param query: 查询对象
        :type query: Query
        :rtype: Query
        """
        dumped = query.dump()
        dumped['className'] = query._query_class._class_name
        self._add_condition(key, '$inQuery', dumped)
        return self

    def does_not_match_query(self, key, query):
        """
        增加查询条件，限制查询结果对象指定字段的值，与另外一个查询对象的返回结果不相同。

        :param key: 查询条件字段名
        :param query: 查询对象
        :type query: Query
        :rtype: Query
        """
        dumped = query.dump()
        dumped['className'] = query._query_class._class_name
        self._add_condition(key, '$notInQuery', dumped)
        return self

    def matched_key_in_query(self, key, query_key, query):
        """
        增加查询条件，限制查询结果对象指定字段的值，与另外一个查询对象的返回结果指定的值相同。

        :param key: 查询条件字段名
        :param query_key: 查询对象返回结果的字段名
        :param query: 查询对象
        :type query: Query
        :rtype: Query
        """
        dumped = query.dump()
        dumped['className'] = query._query_class._class_name
        self._add_condition(key, '$select', {'key': query_key, 'query': dumped})
        return self

    def does_not_match_key_in_query(self, key, query_key, query):
        """
        增加查询条件，限制查询结果对象指定字段的值，与另外一个查询对象的返回结果指定的值不相同。

        :param key: 查询条件字段名
        :param query_key: 查询对象返回结果的字段名
        :param query: 查询对象
        :type query: Query
        :rtype: Query
        """
        dumped = query.dump()
        dumped['className'] = query._query_class._class_name
        self._add_condition(key, '$dontSelect', {'key': query_key, 'query': dumped})
        return self

    def _or_query(self, queries):
        dumped = [q.dump()['where'] for q in queries]
        self._where['$or'] = dumped
        return self

    def _and_query(self, queries):
        dumped = [q.dump()['where'] for q in queries]
        self._where['$and'] = dumped

    def _quote(self, s):
        # return "\\Q" + s.replace("\\E", "\\E\\\\E\\Q") + "\\E"
        return s

    def contains(self, key, value):
        """
        增加查询条件，限制查询结果对象指定最短的值，包含指定字符串。在数据量比较大的情况下会比较慢。

        :param key: 查询条件字段名
        :param value: 需要包含的字符串
        :rtype: Query
        """
        self._add_condition(key, '$regex', self._quote(value))

    def startswith(self, key, value):
        """
        增加查询条件，限制查询结果对象指定最短的值，以指定字符串开头。在数据量比较大的情况下会比较慢。

        :param key: 查询条件字段名
        :param value: 需要查询的字符串
        :rtype: Query
        """
        self._add_condition(key, '$regex', '^' + self._quote(value))
        return self

    def endswith(self, key, value):
        """
        增加查询条件，限制查询结果对象指定最短的值，以指定字符串结尾。在数据量比较大的情况下会比较慢。

        :param key: 查询条件字段名
        :param value: 需要查询的字符串
        :rtype: Query
        """
        self._add_condition(key, '$regex', self._quote(value) + '$')
        return self

    def ascending(self, key):
        """
        限制查询返回结果以指定字段升序排序。

        :param key: 排序字段名
        :rtype: Query
        """
        self._order = [key]
        return self

    def add_ascending(self, key):
        """
        增加查询排序条件。之前指定的排序条件优先级更高。

        :param key: 排序字段名
        :rtype: Query
        """
        self._order.append(key)
        return self

    def descending(self, key):
        """
        限制查询返回结果以指定字段降序排序。

        :param key: 排序字段名
        :rtype: Query
        """
        self._order = ['-{0}'.format(key)]
        return self

    def add_descending(self, key):
        """
        增加查询排序条件。之前指定的排序条件优先级更高。

        :param key: 排序字段名
        :rtype: Query
        """
        self._order.append('-{0}'.format(key))
        return self

    def near(self, key, point):
        """
        增加查询条件，限制返回结果指定字段值的位置与给定地理位置临近。

        :param key: 查询条件字段名
        :param point: 需要查询的地理位置
        :rtype: Query
        """
        self._add_condition(key, '$nearSphere', point)
        return self

    def within_radians(self, key, point, distance):
        """
        增加查询条件，限制返回结果指定字段值的位置在某点的一段距离之内。

        :param key: 查询条件字段名
        :param point: 查询地理位置
        :param distance: 距离限定（弧度）
        :rtype: Query
        """
        self.near(key, point)
        self._add_condition(key, '$maxDistance', distance)
        return self

    def within_miles(self, key, point, distance):
        """
        增加查询条件，限制返回结果指定字段值的位置在某点的一段距离之内。

        :param key: 查询条件字段名
        :param point: 查询地理位置
        :param distance: 距离限定（英里）
        :rtype: Query
        """
        return self.within_radians(key, point, distance / 3958.8)

    def within_kilometers(self, key, point, distance):
        """
        增加查询条件，限制返回结果指定字段值的位置在某点的一段距离之内。

        :param key: 查询条件字段名
        :param point: 查询地理位置
        :param distance: 距离限定（千米）
        :rtype: Query
        """
        return self.within_radians(key, point, distance / 6371.0)

    def within_geo_box(self, key, southwest, northeast):
        """
        增加查询条件，限制返回结果指定字段值的位置在指定坐标范围之内。

        :param key: 查询条件字段名
        :param southwest: 限制范围西南角坐标
        :param northeast: 限制范围东北角坐标
        :rtype: Query
        """
        self._add_condition(key, '$within', {'$box': [southwest, northeast]})
        return self

    def include(self, *keys):
        """
        指定查询返回结果中包含关联表字段。

        :param keys: 关联子表字段名
        :rtype: Query
        """
        if len(keys) == 1 and isinstance(keys[0], (list, tuple)):
            keys = keys[0]
        self._include += keys
        return self

    def select(self, *keys):
        """
        指定查询返回结果中只包含某些字段。可以重复调用，每次调用的包含内容都将会被返回。

        :param keys: 包含字段名
        :rtype: Query
        """
        if len(keys) == 1 and isinstance(keys[0], (list, tuple)):
            keys = keys[0]
        self._select += keys
        return self
